package com.tzld.piaoquan.recommend.server.service.rank.strategy;


import com.alibaba.fastjson.JSONObject;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.feature.domain.video.base.ItemFeature;
import com.tzld.piaoquan.recommend.feature.domain.video.base.RequestContext;
import com.tzld.piaoquan.recommend.feature.domain.video.base.UserFeature;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.common.enums.AppTypeEnum;
import com.tzld.piaoquan.recommend.server.model.MachineInfo;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.remote.FeatureRemoteService;
import com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.RankResult;
import com.tzld.piaoquan.recommend.server.service.rank.RankService;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorFeature;
import com.tzld.piaoquan.recommend.server.service.rank.processor.RankProcessorDensity;
import com.tzld.piaoquan.recommend.server.service.recall.RecallResult;
import com.tzld.piaoquan.recommend.server.service.recall.strategy.*;
import com.tzld.piaoquan.recommend.server.service.score.ScoreParam;
import com.tzld.piaoquan.recommend.server.service.score.ScorerUtils;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.RandomUtils;
import org.apache.commons.lang3.math.NumberUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @author dyp
 */
@Service
@Slf4j
public class RankStrategy4RankModel extends RankService {

    final private String CLASS_NAME = this.getClass().getSimpleName();
    @Override
    public List<Video> mergeAndRankRovRecall(RankParam param) {

        //-------------------地域内部融合-------------------
        List<Video> rovRecallRank = new ArrayList<>();
        rovRecallRank.addAll(extractAndSort(param, RegionHRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionHDupRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, Region24HRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionRelative24HRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, RegionRelative24HDupRecallStrategy.PUSH_FORM));

        removeDuplicate(rovRecallRank);
        rovRecallRank = rovRecallRank.size() <= param.getSize()
                ? rovRecallRank
                : rovRecallRank.subList(0, param.getSize());

        //-------------------地域 sim returnv2 融合-------------------
        rovRecallRank.addAll(extractAndSort(param, SimHotVideoRecallStrategy.PUSH_FORM));
        rovRecallRank.addAll(extractAndSort(param, ReturnVideoRecallStrategy.PUSH_FORM));
        removeDuplicate(rovRecallRank);

        //-------------------排-------------------
        //-------------------序-------------------
        //-------------------逻-------------------
        //-------------------辑-------------------
        List<String> videoIdKeys = rovRecallRank.stream()
                .map(t -> param.getRankKeyPrefix() + t.getVideoId())
                .collect(Collectors.toList());
        List<String> videoScores = redisTemplate.opsForValue().multiGet(videoIdKeys);
        log.info("rank mergeAndRankRovRecall videoIdKeys={}, videoScores={}", JSONUtils.toJson(videoIdKeys),
                JSONUtils.toJson(videoScores));
        if (CollectionUtils.isNotEmpty(videoScores)
                && videoScores.size() == rovRecallRank.size()) {
            for (int i = 0; i < videoScores.size(); i++) {
                rovRecallRank.get(i).setSortScore(NumberUtils.toDouble(videoScores.get(i), 0.0));
            }
            Collections.sort(rovRecallRank, Comparator.comparingDouble(o -> -o.getSortScore()));
        }

        // todo zhangbo 增加排序str模型逻辑



        return rovRecallRank;
    }

    public List<Video> model(List<Video> videos, RankParam param){
        if (videos.isEmpty()){
            return videos;
        }

        Map<String, Object> userFeatureMap = new HashMap<>();
        if (param.getMid() != null && !param.getMid().isEmpty()){
            String midKey = "user_info_4video_" + param.getMid();
            String userFeatureStr = this.redisTemplate.opsForValue().get(midKey);
            if (userFeatureStr != null){
                try{
                    userFeatureMap = JSONUtils.fromJson(userFeatureStr,
                            new TypeToken<Map<String, Object>>() {},
                            userFeatureMap);
                }catch (Exception e){
                    log.error(String.format("parse user json is wrong in {} with {}",
                            this.CLASS_NAME, e));
                }
            }
        }
        log.info("userFeature in model = {}", JSONUtils.toJson(userFeatureMap));

        List<RankItem> rankItems = CommonCollectionUtils.toList(videos, RankItem::new);
        List<Long> videoIds = CommonCollectionUtils.toListDistinct(videos, Video::getVideoId);
        List<String> videoFeatureKeys = videoIds.stream().map(r-> "video_info_" + r)
                .collect(Collectors.toList());
        List<String> videoFeatures = this.redisTemplate.opsForValue().multiGet(videoFeatureKeys);
        if (videoFeatures != null){
            for (int i=0; i<videoFeatures.size(); ++i){
                String vF = videoFeatures.get(i);
                Map<String, Object> vfMap = new HashMap<>();
                if (vF == null){
                    continue;
                }
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, Object>>() {}, vfMap);
                    rankItems.get(i).setFeatureMap(vfMap);
                }catch (Exception e){
                    log.error(String.format("parse video json is wrong in {} with {}",
                            this.CLASS_NAME, e));
                }
            }
        }
        log.info("ItemFeature = {}", JSONUtils.toJson(videoFeatures));

        Map<String, String> sceneFeatureMap =  this.getSceneFeature(param);

        List<RankItem> rovRecallScore = ScorerUtils.getScorerPipeline(ScorerUtils.BASE_CONF)
                .scoring(sceneFeatureMap, userFeatureMap, rankItems);
        log.info("mergeAndRankRovRecallNew rovRecallScore={}", JSONUtils.toJson(rovRecallScore));

        return CommonCollectionUtils.toList(rovRecallScore, i -> {
            // hard code 将排序分数 赋值给video的sortScore
            Video v = i.getVideo();
            v.setSortScore(i.getScore());
            return v;
        });
    }

    private Map<String, String> getSceneFeature(RankParam param) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        sceneFeatureMap.put("ctx_region", param.getProvince());
        sceneFeatureMap.put("ctx_city", param.getCity());
        Calendar calendar = Calendar.getInstance();

        sceneFeatureMap.put("ctx_week", (calendar.get(Calendar.DAY_OF_WEEK) + 6) % 7 + "");
        sceneFeatureMap.put("ctx_hour", new SimpleDateFormat("HH").format(calendar.getTime()));

        return sceneFeatureMap;
    }



}
